(in-package :cl-blog)

(defclass blog-entry ()
  ((category :initarg :category :accessor blog-entry-category)
   (user :initarg :user :accessor blog-entry-user)
   (number :initarg :number :accessor blog-entry-number)
   (title :initarg :title :accessor blog-entry-title)
   (time :initarg :time :accessor blog-entry-time)
   (revised-time :initarg :revised-time :initform 0 :accessor blog-entry-revised-time)
   (contents :initarg :contents :accessor blog-entry-contents)
   (trackbacks :initarg :trackbacks :accessor blog-entry-trackbacks :initform nil)))

(defclass trackback ()
  ((title :initarg :title :accessor trackback-title)
   (url :initarg :url :accessor trackback-url)
   (excerpt :initarg :excerpt :accessor trackback-excerpt)
   (time :initarg :time :accessor trackback-time)
   (blog-name :initarg :blog-name :accessor trackback-blog-name)))

(defvar *blog-entries* nil)

(defvar *blog-entry-counter* 0)

(defmacro with-split-1 ((one rest) split on &body body)
  (let ((one-list (gensym)) (pos (gensym)) (on-binding (gensym)))
    `(let ((,on-binding ,on))
       (multiple-value-bind
             (,one-list ,pos) (split-sequence
                               ,split ,on-binding :count 1)
         (let ((,one (car ,one-list))
               (,rest (subseq ,on-binding ,pos)))
           ,@body)))))

(defun split-whitespace-scrub (str)
  (delete ""
          (split-sequence-if
           #'(lambda (char)
               (member char '(#\space #\tab #\return #\newline))) str)
          :test #'string=))

(defmacro maxf (place &rest forms)
  `(setf ,place (max ,place ,@forms)))

(defmacro with-body-params (param-pairs request &body body)
  (let ((request-name (gensym)))
    `(let* ((,request-name ,request)
            ,@(mapcar #'(lambda (param-pair)
                          `(,(car param-pair)
                             (body-param ,(second param-pair)
                                         (request-body ,request-name))))
                      param-pairs))
       ,@body)))

(defun archives-url (&key category rss)
  (if category
      (concatenate 'string
                   (urlstring (if rss
                                  *rss-url*
                                  *archives-url*))
                   (if rss "?" "/")
                   category)
      (urlstring (if rss *rss-url* *archives-url*))))

(defun make-archives-url (&key category month year)
  (let ((url (copy-url *archives-url*)))
    (setf (url-query url) (format nil "~A~A~A"
                                  (if category
                                      (format nil "category=~A" category)
                                      "")
                                  (if (and category month year)
                                      "&"
                                      "")
                                  (if (and month year)
                                      (format nil "month=~A&year=~A"
                                              month year)
                                      "")))
    (encode-for-pre (urlstring url))))

(deftag :rss-feed-link (stream)
  (html-stream stream
               `((link :rel "alternate" :type "application/rss+xml" :title
                       ,(format nil "RSS Feed for ~A" *blog-short-name*)
                       :href ,(archives-url :rss t)))))

(deftag :sidebar-blog-links (stream)
  (html-stream
   stream
   `(ul
     (li ((a :href ,(urlstring *blog-url-root*))
          "Main"))
     (li ((a :href ,(urlstring *new-entry-url*))
          "New entry"))
     (li ((a :href ,(urlstring *archives-url*))
          "Archives"))
     (li ((a :href ,(archives-url :rss t))
          "Syndicate (RSS)"))
     (li (b ((a :href ,(urlstring *email-redirect-url*))
             "Send Comments"))))))

(deftag :recent-entry-links (stream)
  (html-stream
   stream
   `(ul
     ,@(loop for i from 1 to 10
          for j in *blog-entries*
          collect `(li
                    ((a :href ,(entry-link j))
                     ,(blog-entry-title j)))))))

(deftag :category-links (stream)
  (html-stream
   stream
   `(ul
     ,@(loop for i in *categories*
          collect `(li ((a :href ,(make-archives-url :category i))
                        ,i) " "
                        ((a :href ,(archives-url :category i :rss t))
                         "(RSS)")
                        )))))

(deftag :blog-short-name (stream)
  (write-string *blog-short-name* stream))

(defun blog-wrap-page (stream title body)
  (let-tags ((:page-title (stream) (write-string title stream))
             (:page-body (stream)
                         (html-stream stream
                                      `((div :id "content")
                                        ,@body))))
	    (expand-template-body *blog-template* stream)
	    t))

(defun blog-entry-path (entry)
  (merge-pathnames (prin1-to-string (blog-entry-number entry))
                   *entry-path*))

(defun make-trackback-constructor (trackback)
  (with-slots (title url excerpt time blog-name) trackback
    `(make-instance 'trackback
                    :title ,title
                    :url ,url
                    :excerpt ,excerpt
                    :time ,time
                    :blog-name ,blog-name)))

(defun blog-entry-write-to-file (entry)
  (ensure-directories-exist (blog-entry-path entry))
  (with-open-file (file (blog-entry-path entry)
                        :direction :output
                        :if-exists :supersede)
    (with-slots (category user number title time revised-time contents trackbacks) entry
      (let ((*package* (find-package #.(package-name *package*))))
	(prin1
	 `(push
	   (make-instance 'blog-entry
			  :category ,category
			  :user ,user
			  :number (prog1 ,number (maxf *blog-entry-counter* ,number))
			  :title ,title
			  :time ,time
			  :revised-time ,revised-time
			  :contents ,contents
			  :trackbacks (list ,@(mapcar #'make-trackback-constructor trackbacks)))
	   *blog-entries*) file)))))

(defun delete-blog-entry (entry)
  (setf *blog-entries* (remove entry *blog-entries*))
  (setf *blog-entry-counter*
        (apply #'max 0
               (mapcar #'blog-entry-number *blog-entries*)))
  (delete-file (blog-entry-path entry)))

(defun load-blog-entries ()
  (setf *blog-entries* nil)
  (let ((*package* (find-package #.(package-name *package*))))
    (mapcar #'load (directory (make-pathname :name :wild :defaults *entry-path*))))
  (setf *blog-entries* (sort *blog-entries* #'> :key #'blog-entry-number)))

(defun start-blog ()
  (ignore-errors (stop-listening *listener*))
  (load-blog-entries)
  (start-listening *listener*))

(defvar *authenticated-user*)

(defun new-blog-entry (&key category title contents)
  (incf *blog-entry-counter*)
  (let ((new-entry (make-instance 'blog-entry :number *blog-entry-counter*
                                  :category category
                                  :user *authenticated-user*
                                  :title title
                                  :time (get-universal-time)
                                  :contents contents)))
    (push new-entry *blog-entries*)
    (blog-entry-write-to-file new-entry)
    new-entry))

(defun change-blog-entry (entry &key (category (blog-entry-category entry))
                          (title (blog-entry-title entry))
                          (contents (blog-entry-contents entry)))
  (setf (blog-entry-revised-time entry)
        (get-universal-time))
  (setf (blog-entry-category entry) category)
  (setf (blog-entry-title entry) title)
  (setf (blog-entry-contents entry) contents)
  (blog-entry-write-to-file entry))

(defclass validate-unhandled-part-mixin ()
  ((validate-lambda :initform (constantly t)
                    :initarg :validate-lambda
                    :accessor validate-lambda)))

(defclass main-handler (handler) ())

(defclass authenticate-mixin () ())

(defclass new-entry-handler (handler authenticate-mixin) ())

(defun entry-from-regexp-validator (regexp)
  (lambda (unhandled-part)
    (let* ((l (nth-value 1 (cl-ppcre:scan-to-strings
                            regexp unhandled-part)))
           (str (and l (elt
                        l
                        0))))
      (and str (find-entry (parse-integer str))))))

(defclass display-entry-handler (handler validate-unhandled-part-mixin)
  ((validate-lambda :initform (entry-from-regexp-validator "^(\\d+)$"))))

(defclass edit-entry-handler (handler validate-unhandled-part-mixin authenticate-mixin)
  ((validate-lambda :initform (entry-from-regexp-validator "^(\\d+)$"))))

(defclass trackback-handler (handler validate-unhandled-part-mixin)
  ((validate-lambda :initform (entry-from-regexp-validator "^(\\d+)$"))))

(defclass delete-entry-handler (handler validate-unhandled-part-mixin authenticate-mixin)
  ((validate-lambda :initform (entry-from-regexp-validator "^(\\d+)$"))))

(defun category-from-regexp-validator (regexp)
  (lambda (unhandled-part)
    (or (eql (length unhandled-part) 0)
        (let ((str (elt
                    (nth-value 1 (cl-ppcre:scan-to-strings
                                  regexp unhandled-part))
                    0)))
          (and str (find str *categories* :test #'string-equal))))))
  
(defclass archives-handler (handler #+nil validate-unhandled-part-mixin)
  (#+nil
   (validate-lambda :initform (category-from-regexp-validator "^/(\\w+)$"))))

(defclass rss-handler (handler validate-unhandled-part-mixin)
  ((validate-lambda :initform (category-from-regexp-validator "^\\?(\\w+)$"))))

(defun send-doctype (request)
  (write-string "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\" \"http://www.w3.org/TR/html4/loose.dtd\">" (request-stream request))
  (terpri (request-stream request)))

(defmethod handle-request-response ((handler rss-handler) method request)
  (request-send-headers request :expires 0 :content-type "text/xml")
  (let* ((unhandled-part (request-unhandled-part request))
         (category (if (eql (length unhandled-part) 0)
                       nil
                       (elt
                        (nth-value 1 (cl-ppcre:scan-to-strings
                                      "^\\?(\\w+)$" unhandled-part))
                        0))))
    (format (request-stream request)
            "<?xml version=\"1.0\"?>~%")
    (html-stream
     (request-stream request)
     `((|rss| :|version| "2.0")
       (|channel|
        (|title| ,*blog-short-name*)
        (|link| ,(urlstring *blog-url-root*))
        (|description| ,(if category
                            (format nil "~A archives for category ~A"
                                    *blog-short-name*
                                    category)
                            (format nil "~A archives"
                                    *blog-short-name*)))
        (|language| "en")
        ,@(loop for i in *blog-entries*
             for j from 1 to 20
             if (or (not category)
                    (string-equal (blog-entry-category i) category))
             collect
             `(|item|
               (|title|
                ,(blog-entry-title i))
               (|link|
                ,(entry-link i))
               (|pubDate|
                ,(universal-time-to-rfc-date
                  (blog-entry-time i)))
               (|description|
                ,(encode-for-http
                  (encode-for-pre
                   (blog-entry-contents i)))))))))))

(defparameter *week-day-names* #("Mon" "Tue" "Wed" "Thu" "Fri" "Sat" "Sun"))
(defparameter *month-names* #("Jan" "Feb" "Mar" "Apr" "May" "Jun" "Jul"
			      "Aug" "Sep" "Oct" "Nov" "Dec"))

(defun date-name-from-number (number name-mapping)
  (elt name-mapping (mod (1- number) (length name-mapping))))

(defun format-entry-date (date)
  (with-date date nil
             (format nil "~A, ~A ~A ~A"
                     (elt *week-day-names* day-of-week)
                     day-of-month
                     (date-name-from-number month *month-names*)
                     year)))

(defun format-month (date)
  (with-date date nil
             (format nil "~A ~A"
                     (date-name-from-number month *month-names*)
                     year)))

(defun beginning-of-month (date)
  (multiple-value-bind (second minute hour date month year)
      (decode-universal-time date)
    (declare (ignore second minute hour date))
    (encode-universal-time 0 0 0 1 month year)))

(defun date-beginning-month (month year &key next-month)
  (if next-month
      (encode-universal-time 0 0 0 1 (1+ (mod month 12)) (if (eql month 12)
                                                             (1+ year)
                                                             year))
      (encode-universal-time 0 0 0 1 month year)))

(defun month-and-year-from-strings-or-current (month year)
  (let* ((actual-month (if month (parse-integer month :junk-allowed t) 0))
         (actual-year (if year (parse-integer year :junk-allowed t) 0))
         (month (if (> actual-month 0) actual-month
                    (with-date (get-universal-time) nil month)))
         (year (if (> actual-year 0) actual-year
                   (with-date (get-universal-time) nil year))))
    (values month year)))

(defun month-selector-div (category)
  (let ((first-entry (car (last *blog-entries*))))
    (if first-entry
        `((div :class "archives-box")
          (ul
           ,@(loop with first-entry-year = (with-date (blog-entry-time first-entry) nil year)
                for i from first-entry-year
                to (with-date (get-universal-time) nil year)
                appending (loop for j from (if (eql i first-entry-year)
                                               (with-date (blog-entry-time first-entry) nil month) 1)
                             to 12
                             if (> (count-if #'(lambda (ent)
                                                 (>= (date-beginning-month j i :next-month t)
                                                     (blog-entry-time ent)
                                                     (date-beginning-month j i)))
                                             *blog-entries*) 0)
                             collect `(li
                                       ((a :href ,(make-archives-url :category category
                                                                     :month j
                                                                     :year i))
                                        ,(format-month (date-beginning-month j i))))))))
        `((div :class "archives-box")))))

(defmethod handle-request-response ((handler archives-handler) method request)
  (let* ((category (car (url-query-param (request-url request) "category")))
         (month-string (car (url-query-param (request-url request) "month")))
         (year-string (car (url-query-param (request-url request) "year"))))
    (multiple-value-bind (month year)
        (month-and-year-from-strings-or-current month-string year-string)
      (let ((bom (date-beginning-month month year))
            (eom (date-beginning-month month year :next-month t)))
        (request-send-headers request :expires 0
                              :content-type "text/html; charset=iso-8859-1")
        (send-doctype request)
        (blog-wrap-page
         (request-stream request)
         (if category
             (format nil "~A Archives: ~A" category
                     (format-month bom))
             (format nil "Archives for ~A" (format-month bom)))
         `(,(month-selector-div category)
            ,@(nreverse
               (loop for i in *blog-entries*
                  if (and (or (not category)
                              (string-equal (blog-entry-category i) category))
                          (>= eom (blog-entry-time i) bom))
                  collect (format-entry i)))))))))

(defmethod handle-request-response :around ((handler validate-unhandled-part-mixin) method request)
  (if (funcall (validate-lambda handler) (request-unhandled-part request))
      (call-next-method)
      (progn
        (request-send-headers request :expires 0
                              :content-type "text/html; charset=iso-8859-1")
        (send-doctype request)
        (blog-wrap-page (request-stream request) "No page for you!" nil))))

(defun find-entry (number)
  (find number *blog-entries* :key #'blog-entry-number))

(defun entry-link (entry)
  (urlstring (merge-url *display-entry-url*
                        (prin1-to-string (blog-entry-number entry)))))

(defun entry-trackback (entry)
  (urlstring (merge-url *trackback-url*
                        (prin1-to-string (blog-entry-number entry)))))

(defun edit-entry-link (entry)
  (urlstring
   (merge-url
    *edit-entry-url*
    (prin1-to-string (blog-entry-number entry)))))

(defun delete-entry-link (entry)
  (urlstring
   (merge-url
    *delete-entry-url*
    (prin1-to-string (blog-entry-number entry)))))

(defun format-entry-user (entry)
  (let ((user (blog-entry-user entry)))
    (if (member user *users* :test #'string-equal)
        (let* ((user-keyword (intern (string-upcase user) :keyword))
               (name (configuration-values *user-configuration*
                                           :real-name user-keyword :users))
               (link (format nil "~A?user=~A"
                             (urlstring *email-redirect-url*)
                             user)))
          (if name
              (if link
                  `((a :href ,link) ,(first name))
                  name)
              user))
        user)))

(defun format-entry (entry &key (display-link t) (display-actions t))
  `((div :class "entry")
    ,@(if display-actions
          (list (format nil "<!--
<rdf:RDF xmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"
         xmlns:dc=\"http://purl.org/dc/elements/1.1/\"
         xmlns:trackback=\"http://madskills.com/public/xml/rss/module/trackback/\">
<rdf:Description
    rdf:about=\"~A\"
    dc:identifier=\"~A\"
    dc:title=\"~A\"
    trackback:ping=\"~A\" />
</rdf:RDF>
-->
" (entry-link entry) (entry-link entry) (blog-entry-title entry) (entry-trackback entry))))
    ((div :class "entry-head")
     (h2 
      ,(blog-entry-title entry))
     ((div :class "entry-data")
      "Posted on "
      ,(format-entry-date (blog-entry-time entry))
      " in "
      ((a :href ,(make-archives-url :category (blog-entry-category entry)))
       ,(blog-entry-category entry))
      " by "
      ,(format-entry-user entry)
      ,@(if (> (blog-entry-revised-time entry)
	       (blog-entry-time entry))
	    `(,(format nil
		       "<br>Last edited ~A"
		       (format-entry-date (blog-entry-revised-time entry)))))))
    ((div :class "entry-text")
     ,(blog-entry-contents entry))
    ,@(when display-actions
            `(((div :class "entry-footer")
               ,@(if display-link
                     `(((a :class "entry-footer" :href ,(entry-link entry))
                        ,(if (blog-entry-trackbacks entry)
                             (format nil "Permanent Link (~A)"
                                     (length (blog-entry-trackbacks entry)))
                             "Permanent Link"))))
               " | "
               ((a :class "entry-footer" :href ,(edit-entry-link entry))
                "Edit")
               " | "
               ((a :class "entry-footer" :href ,(delete-entry-link entry))
                "Delete"))))))

(defun format-trackback (trackback)
  `((div :class "entry")
    ((div :class "entry-head")
     ((:a :href ,(trackback-url trackback))
      ,(encode-for-pre (trackback-title trackback)))
     ((div :class "entry-data")
      "From "
      (i ,(encode-for-pre (trackback-blog-name trackback)))
      " on " ,(format-entry-date (trackback-time trackback))
      ))
    ((div :class "entry-text")
     ,(encode-for-pre (trackback-excerpt trackback)))
    ))

(defmethod handle-request-response ((handler main-handler) method request)
  (request-send-headers request :expires 0
                        :content-type "text/html; charset=iso-8859-1")
  (send-doctype request)
  (blog-wrap-page
   (request-stream request)
   (format nil "Blog Entries")
   (loop for i from 1 to 10
      for j in *blog-entries*
      collecting (format-entry j))))

(defun blog-entry-form (post-url &key (title "") (category "") (contents "") (number "") (submit-button-text "Post entry"))
  `((div :id "entry-form")
    ((form :method post :action ,post-url)
     ((input :type hidden :name "number" :value
             ,number))
     ((table :class "new-entry-form")
      (tr
       (th "Title: ")
       (td ((input :type text :name "title"
                   :size 40
                   :value
                   ,(encode-for-argument
                     title)))))
      (tr
       (th "Category: ")
       (td ((select :name "category")
            ,@(mapcar #'(lambda (cat)
                          `((option :value ,cat
                                    ,@(if (string-equal cat
                                                        category)
                                          '(:selected "SELECTED"))
                                    ) ,cat))
                      *categories*))))
      (tr
       ((th :valign "top") "Contents: ")
       (td ((textarea :rows 24 :cols 40 :name "contents")
            ,(encode-for-pre
              contents))))
      (tr
       (td)
       (td ((input :type submit :value
                   ,(encode-for-pre submit-button-text)))
           " "
           ((input :type submit :name "preview" :value
                   "Preview"))))))))

(defun copy-entry (entry)
  (make-instance 'blog-entry
                 :contents (blog-entry-contents entry)
                 :user (blog-entry-user entry)
                 :title (blog-entry-title entry)
                 :time (blog-entry-time entry)
                 :revised-time (blog-entry-revised-time entry)
                 :category (blog-entry-category entry)
                 :number (blog-entry-number entry)))

(defmethod handle-request-response ((handler edit-entry-handler) (method (eql :get)) request)
  (let ((entry (find-entry (parse-integer (request-unhandled-part request)))))
    (request-send-headers request :expires 0
                          :content-type "text/html; charset=iso-8859-1")
    (send-doctype request)
    (blog-wrap-page
     (request-stream request)
     "Edit Blog Entry"
     (list (blog-entry-form
            (urlstring
             (merge-url
              *edit-entry-url*
              (request-unhandled-part request)))
            :number (prin1-to-string (blog-entry-number entry))
            :title (blog-entry-title entry)
            :category (blog-entry-category entry)
            :contents (blog-entry-contents entry)
            :submit-button-text "Edit entry"))
     )))

(defmethod handle-request-response ((handler delete-entry-handler) (method (eql :get)) request)
  (let ((entry (find-entry (parse-integer (request-unhandled-part request)))))
    (request-send-headers request :expires 0
                          :content-type "text/html; charset=iso-8859-1")
    (send-doctype request)
    (blog-wrap-page
     (request-stream request)
     "Confirm Delete Blog Entry"
     (list `((form :method "post" :action ,(urlstring (request-url request)))
             ((span :class "small-header")
              "Are you sure you want to delete this entry?")
             ,(format-entry entry)
             ((input :type submit :name "confirm" :value "Yes"))
             ((input :type submit :name "noconfirm" :value "No")))
           ))))

(defmethod handle-request-response ((handler delete-entry-handler) (method (eql :post)) request)
  (let ((entry (find-entry (parse-integer (request-unhandled-part request)))))
    (with-body-params ((confirm "confirm"))
      request
      (request-send-headers request :expires 0
                            :content-type "text/html; charset=iso-8859-1")
      (send-doctype request)
      (if (equalp confirm "yes")
          (progn
            (delete-blog-entry entry)
            (blog-wrap-page
             (request-stream request)
             "Deleted!" nil))
           (blog-wrap-page
            (request-stream request)
            "Not deleted!" nil)))))

(defmethod handle-request-response ((handler edit-entry-handler) (method (eql :post)) request)
  (with-body-params ((title "title")
                     (category "category")
                     (contents "contents")
                     (number "number")
                     (preview "preview"))
    request
    (let ((entry (find-entry (parse-integer number))))
      (request-send-headers request :expires 0
                            :content-type "text/html; charset=iso-8859-1")
      (send-doctype request)
      (if (equalp preview "Preview")
          (let ((entry (copy-entry entry)))
            (setf (blog-entry-contents entry) contents)
            (setf (blog-entry-category entry) category)
            (setf (blog-entry-title entry) title)
            (blog-wrap-page
             (request-stream request)
             "Preview edit"
             (list
              (format-entry entry :display-actions nil)
              '(p)
              (blog-entry-form
               (urlstring
                (merge-url
                 *edit-entry-url*
                 (request-unhandled-part request)))
               :number (prin1-to-string (blog-entry-number entry))
               :title title
               :category category
               :contents contents
               :submit-button-text "Edit entry"))))
          (if (not (find 0 (list title category contents) :key #'length))
              (progn
                (change-blog-entry entry :title title :category category :contents contents)
                (blog-wrap-page
                 (request-stream request)
                 "Changed"
                 '(((div :id "entry-form")
                    (h2 "Your entry has been modified.")))))
               (blog-wrap-page
                (request-stream request)
                "Not changed"
                '(((div :id "entry-form")
                   (h2 "Please fill in all fields!")))))))))

(defmethod handle-request-response ((handler new-entry-handler) (method (eql :get)) request)
  (request-send-headers request :expires 0
                        :content-type "text/html; charset=iso-8859-1")
  (send-doctype request)
  (blog-wrap-page
   (request-stream request)
   "New Blog Entry"
   (list (blog-entry-form (urlstring *new-entry-url*)))))

(defun md5-from-string (md5sum)
  (loop for i below (- (length md5sum) 1) by 2
     collect (parse-integer md5sum :start i :end (+ i 2) :radix 16)))

(defmethod handle-request-response :around ((handler authenticate-mixin) method request)
  (let ((auth (car (request-header request :authorization)))
        (authenticated nil)
        (authed-user nil))
    (if auth
        (let* ((base64-auth (second (split-whitespace-scrub auth)))
               (de-base64 (base64-string-to-string base64-auth)))
          (with-split-1 (user pass) #\: de-base64
                        (if (and
                             (member user *users* :test #'string-equal)
                             (equal (coerce
                                     (md5sum-sequence pass) 'list)
                                    (md5-from-string
                                     (configuration-value *user-configuration*
                                                          :password (intern (string-upcase user) :keyword) :users))))
                            (setf
                             authed-user user
                             authenticated t)))))
    ;;(format t "auth is ~S ~S~%" auth authenticated)
    (if authenticated
        (let ((*authenticated-user* authed-user))
          (call-next-method))
        (progn
          (request-send-headers request
                                :content-type "text/html; charset=iso-8859-1"
                                :expires 0
                                :response-code 401
                                :response-text "Unauthorized"
                                :www-authenticate
                                (format nil "Basic realm=\"~A\""
                                        *blog-short-name*))
          (send-doctype request)
          (html-stream
           (request-stream request)
           `(html (body (h1 "No page for you!"))))))))

(defun search-for-trackbacks-for-text (text)
  (let ((url-regexp "(?i)(http://[^ \"]+)([ .,)!?\"]|$)"))
    (loop for (match new-begin) = (multiple-value-list (cl-ppcre:scan url-regexp text))
         for url = nil
         while match
         do (let ((got-url (elt (nth-value 1 (cl-ppcre:scan-to-strings url-regexp text)) 0)))
              (setf text (subseq text new-begin))
              (setf url got-url))
         if url collect url)))

(defun strip-html (text)
  (cl-ppcre:regex-replace-all "<(\"[^\"]+\"|[^>\"])*>" text ""))

(defun autotrackback-entry (entry)
  (let ((urls (search-for-trackbacks-for-text (blog-entry-contents entry))))
    (loop for url in urls
         for ping = (trackback:autodetect-ping-for-url url)
         if ping
         do (trackback:ping ping :title (urlstring-escape (blog-entry-title entry))
                            :url (urlstring-escape (entry-link entry))
                            :excerpt (urlstring-escape (strip-html (blog-entry-contents entry)))
                            :blog-name *blog-short-name*)
         if ping
         collect url)))

(defmethod handle-request-response ((handler new-entry-handler) (method (eql :post)) request)
  (with-body-params ((title "title")
                     (category "category")
                     (contents "contents")
                     (preview "preview"))
    request
    (request-send-headers request :expires 0
                          :content-type "text/html; charset=iso-8859-1")
    (send-doctype request)
    (let ((entry (make-instance 'blog-entry
                                :contents contents
                                :category category
                                :user *authenticated-user*
                                :title title
                                :time (get-universal-time))))
      (if (equalp preview "Preview")
          (blog-wrap-page
           (request-stream request)
           "Preview entry"
           (list
            (format-entry entry :display-actions nil)
            '(p)
            (blog-entry-form
             (urlstring
              *new-entry-url*)
             :title title
             :category category
             :contents contents
             )))
	   (if (not (find 0 (list title category contents) :key #'length))
	       (let* ((entry
                       (new-blog-entry :title title :category category :contents contents))
                      (urls-pinged (ignore-errors (autotrackback-entry entry))))
                   (blog-wrap-page (request-stream request) "Posted"
                                   `(((div :id "entry-form")
                                      (h2 "Your new entry has been posted.")
				      ((div :class "trackback")
				       ,@(when urls-pinged
					   `("Trackback pings have been sent to the following URLs:"
					     (ul
					      ,@(mapcar #'(lambda (url)
							    `(li ((a :href ,url) ,url)))
							urls-pinged)))))
                                              ))))
	       (blog-wrap-page (request-stream request) "Not posted"
			       '(((div :id "entry-form")
                                  (h2 "Please fill in all fields!")))))))))

(defmethod handle-request-response ((handler trackback-handler) (method (eql :post)) request)
  (with-body-params ((title "title")
                     (url "url")
                     (excerpt "excerpt")
                     (blog-name "blog_name"))
    request
    (request-send-headers request :expires 0
                          :content-type "text/xml; charset=iso-8859-1")
    (let ((entry 
           (let ((str (elt (nth-value 1 (cl-ppcre:scan-to-strings
                                         "^(\\d+)$" (request-unhandled-part request))) 0)))
             (and str (find-entry (parse-integer str))))))
      (if (not entry)
          (format (request-stream request)
                  "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>
<response>
<error>1</error>
<message>The specified blog entry could not be found.</message>
</response>
")
          (if (string-equal title blog-name)
              ;; hackish antispam
              (signal 'http-not-found)
              (let ((trackback (make-instance 'trackback
                                              :title title :url url
                                              :excerpt excerpt :time (get-universal-time)
                                              :blog-name blog-name)))
                (setf (blog-entry-trackbacks entry)
                      (nconc (blog-entry-trackbacks entry) (list trackback)))
                (blog-entry-write-to-file entry)
                (format (request-stream request)
                        "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>
<response>
<error>0</error>
</response>
"))))
          t)))

(defun despam-trackbacks ()
  (loop for entry in *blog-entries*
       do (progn
            (setf (blog-entry-trackbacks entry)
                  (remove-if #'(lambda (e)
                                 (string-equal (trackback-title e)
                                               (trackback-blog-name e)))
                             (blog-entry-trackbacks entry)))
            (blog-entry-write-to-file entry))))

(defmethod handle-request-response ((handler display-entry-handler) method request)
  (let ((entry 
         (let ((str (elt (nth-value 1 (cl-ppcre:scan-to-strings
                                       "^(\\d+)$" (request-unhandled-part request))) 0)))
           (and str (find-entry (parse-integer str))))))
    (request-send-headers request :expires 0
                          :content-type "text/html; charset=iso-8859-1")
    (send-doctype request)
    (blog-wrap-page
     (request-stream request)
     (blog-entry-title entry)
     (list* (format-entry entry)
            `((div :class "trackback") "Trackback pings for this entry are listed below. The URL to ping for this entry is: " (b ,(entry-trackback entry)))
            (mapcar #'format-trackback (blog-entry-trackbacks entry))))))

(defclass email-redirect-handler (handler) ())

(defmethod handle-request-response ((handler email-redirect-handler) method request)
  (let* ((user-to-email (car (url-query-param (request-url request) "user")))
         (user-keyword (if user-to-email (intern (string-upcase user-to-email) :keyword)))
         (link (if user-keyword (configuration-values *user-configuration*
                                                      :link user-keyword :users))))
    (let ((email-url (if link (first link)
                         (concatenate 'string "mailto:" *owner-email*))))
      (request-send-headers
       request
       :location email-url
       :expires "Fri, 30 Oct 1998 14:19:41 GMT"
       :pragma "no-cache"
       :response-code 302 :response-text "Redirected")
      (html-stream
       (request-stream request)
       `(html (body (h1 "Redirected"))))
      t)))

(install-handler (http-listener-handler *listener*)
                 (make-instance 'main-handler)
                 (urlstring *blog-url-root*) t)

(install-handler (http-listener-handler *listener*)
                 (make-instance 'new-entry-handler)
                 (urlstring *new-entry-url*) t)

(install-handler (http-listener-handler *listener*)
                 (make-instance 'display-entry-handler)
                 (urlstring *display-entry-url*) nil)

(install-handler (http-listener-handler *listener*)
                 (make-instance 'archives-handler)
                 (urlstring *archives-url*) nil)

(install-handler (http-listener-handler *listener*)
                 (make-instance 'rss-handler)
                 (urlstring *rss-url*) nil)

(install-handler (http-listener-handler *listener*)
                 (make-instance 'edit-entry-handler)
                 (urlstring *edit-entry-url*) nil)

(install-handler (http-listener-handler *listener*)
                 (make-instance 'delete-entry-handler)
                 (urlstring *delete-entry-url*) nil)

(install-handler (http-listener-handler *listener*)
                 (make-instance 'email-redirect-handler)
                 (urlstring *email-redirect-url*) nil)

(install-handler (http-listener-handler *listener*)
                 (make-instance 'trackback-handler)
                 (urlstring *trackback-url*) nil)
